// 03-01-04 Cleaned Up
// Texture Loader
// Based on Jan Horns texture loader unit.
// Currently loads
// - 24bit bmp files
// - 24bit tga files
// - 32bit tga files (alpha channel)

unit Textures;

interface

uses
 DglOpengl, Classes;

function LoadTexture(Filename: String; var Texture: GLuInt): Boolean;
function LoadTextureStream(Stream: Tstream;Kind: String; var Texture: GLuInt): Boolean;

implementation


uses
 Windows, SysUtils;

type TTextureImage = record										// Structure Name
	imageData : PChar;										// Image Data (Up To 32 Bits)
	bpp : GLuint;											// Image Color Depth In Bits Per Pixel.
	width : GLuint;											// Image Width
	height : GLuint;											// Image Height
	texID : GLuint;
     end;

{------------------------------------------------------------------}
{  Create the Texture                                              }
{------------------------------------------------------------------}
function CreateTexture(Width, Height, Format : GluInt; pData : Pointer) : Integer;
var
 Texture : GluInt;
begin
 glGenTextures(1, @Texture);
 glBindTexture(GL_TEXTURE_2D, Texture);
 glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);  //Texture blends with object background
 //  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL); //Texture does NOT blend with object background

 { Select a filtering type. BiLinear filtering produces very good results with little performance impact
   GL_NEAREST               - Basic texture (grainy looking texture)
   GL_LINEAR                - BiLinear filtering
   GL_LINEAR_MIPMAP_NEAREST - Basic mipmapped texture
   GL_LINEAR_MIPMAP_LINEAR  - BiLinear Mipmapped texture
 }

 glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR); //only first two can be used
 glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR); //all of the above can be used

 if Format = GL_RGBA then
  gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA, Width, Height, GL_RGBA, GL_UNSIGNED_BYTE, pData)
 else
  gluBuild2DMipmaps(GL_TEXTURE_2D, 3, Width, Height, format, GL_UNSIGNED_BYTE, pData);
  //glTexImage2D(GL_TEXTURE_2D, 0, 3, Width, Height, 0, GL_RGB, GL_UNSIGNED_BYTE, pData);  // Use when not wanting mipmaps to be built by openGL

 result :=Texture;
end;

function LoadBMPTexture(Filename: Tstream; var Texture: GLuint): Boolean;
var
 FileHeader: TBITMAPFILEHEADER;
 InfoHeader: TBITMAPINFOHEADER;
 Palette: array of RGBQUAD;
 BitmapLength: LongWord;
 PaletteLength: LongWord;
 ReadBytes: LongWord;
 TextureImage:  TTextureImage;
begin
 LoadBMPTexture:=False;
 // Get header information
 {ReadBytes:=}Filename.Read(FileHeader, SizeOf(FileHeader));
 {ReadBytes:=}Filename.Read(InfoHeader, SizeOf(InfoHeader));

 // Get palette
 PaletteLength := InfoHeader.biClrUsed;
 SetLength(Palette, PaletteLength);
 ReadBytes:=Filename.Read(Palette,PaletteLength);
 if (ReadBytes <> PaletteLength) then
 begin
  MessageBox(0, PChar('Error reading palette'), PChar('BMP Loader'), MB_OK);
  Exit;
 end;

 textureimage.Width := InfoHeader.biWidth;
 textureimage.Height := InfoHeader.biHeight;
 BitmapLength := InfoHeader.biSizeImage;
 if BitmapLength = 0 then
  BitmapLength := Textureimage.Width * Textureimage.Height * InfoHeader.biBitCount Div 8;

 // Get the actual pixel data
 GetMem(textureimage.imageData, BitmapLength);
 ReadBytes:=Filename.Read(textureimage.imagedata^, BitmapLength);
 if (ReadBytes <> BitmapLength) then
 begin
  MessageBox(0, PChar('Error reading bitmap data'), PChar('BMP Unit'), MB_OK);
  Exit;
 end;

 //BMP files are stored BGR so use GL_BGR
 Texture:=CreateTexture(textureimage.width, textureimage.height, GL_BGR, textureimage.imagedata);

 //clean up
 FreeMem(textureimage.imageData)
end;

function LoadTGATexture(filename: Tstream; var Texture : GLuint): Boolean;
const
 TGAheader : array [0..11] of GLubyte = (0,0,2,0,0,0,0,0,0,0,0,0);	// Uncompressed TGA Header
 TGAComheader : array [0..11] of GLubyte = (0,0,10,0,0,0,0,0,0,0,0,0);	// Compressed TGA Header
var
 gltype : GLuint;								// Set The Default GL Mode To RBGA (32 BPP)
 Compressed:boolean;
 TGAcompare : array [0..11] of GLubyte;								// Used To Compare TGA Header
 header : array [0..5] of GLubyte;									// First 6 Useful Bytes From The Header
 bytesPerPixel : GLuint;					// Holds Number Of Bytes Per Pixel Used In The TGA File
 imageSize : GLuint;									// Used To Store The Image Size When Setting Aside Ram
 i : GLuint;										// Temporary Variable
 Tm: Char;
 TextureImage:  TTextureImage;
// tgafile : integer;
 PixelCount, CurrentPixel, CurrentByte: gluInt;
 ColorBuffer: Pchar;
 ChunkHeader: gluByte;
 Counter, Ret: GLuint;
begin
 result:=False;
// gltype:= GL_RGBA;
// compressed:=false;

 //check if file is a tga file
 Ret := Filename.Read(TGAcompare,sizeof(TGAcompare));
 if (Ret <> sizeof(TGAcompare)) then
 begin
  MessageBox(0, PChar('Not a TGA file'), PChar('TGA Loader'), MB_OK);
  result := false;									// Return False
  exit;
 end;

 //Determine if the tga file is compressed or not
 if (CompareMem(@TGAheader, @TGAcompare, sizeof(TGAheader)) = false) then
 begin
  if (CompareMem(@TGAComheader, @TGAcompare, sizeof(TGAComheader)) = True) then
  Begin
   Compressed := True;
  end
  else
  Begin
//   i := 0;
   MessageBox(0, PChar('Not a TGA file'), PChar('TGA Loader'), MB_OK);
   result := false;									// Return False
   exit;
  end;
 end
 else
 begin
  Compressed := False;
 end;

 Ret := Filename.Read(header, sizeof(header));
 if (Ret <> sizeof(header)) then
 begin
     //ShowError(10, 'TGATexture.LoadTGA', False);
     //if (tgafile = -1) then									// Did The File Even Exist? *Added Jim Strong*
     //begin
     //  result := false;									// Return False
     //  exit;
     //end
     //else
     //begin
     //  fileclose(tgafile);									// If Anything Failed, Close The File
       result := false;									// Return False
       exit;
     //end;
   end;

  textureimage.width  := header[1] * 256 + header[0];			// Determine The TGA Width	(highbyte*256+lowbyte)
  textureimage.height := header[3] * 256 + header[2];			// Determine The TGA Height	(highbyte*256+lowbyte)

  if (textureimage.width <= 0)	or								// Is The Width Less Than Or Equal To Zero
     (textureimage.height <= 0)	or								// Is The Height Less Than Or Equal To Zero
     ((header[4] <> 24) and (header[4] <> 32)) then					// Is The TGA 24 or 32 Bit?
  begin
    //ShowError(11, 'TGATexture.LoadTGA', False);
//    fileclose(tgafile);										// If Anything Failed, Close The File
    result := false;										// Return False
    exit;
  end;

  textureimage.bpp	:= header[4];							// Grab The TGA's Bits Per Pixel (24 or 32)
  bytesPerPixel	:= textureimage.bpp div 8;						// Divide By 8 To Get The Bytes Per Pixel
  imageSize	:= textureimage.width * textureimage.height * bytesPerPixel;	// Calculate The Memory Required For The TGA Data


  GetMem(textureimage.imageData, imageSize);		// Reserve Memory To Hold The TGA Data
  //Setlength(texture.imagedata, imagesize);
  if Not Compressed then
  begin
   Ret:=FileName.Read(textureimage.imagedata^, integer(Imagesize));
   //Ret := fileread(tgafile, textureimage.imageData^, integer(imageSize));
   if (textureimage.imageData = nil) or	       // Does The Storage Memory Exist?
      (Ret <> imageSize)	// Does The Image Size Match The Memory Reserved?
      then
      begin
        if (textureimage.imageData <> nil)						// Was Image Data Loaded
           then freemem(textureimage.imageData);						// If So, Release The Image Data

        //fileclose(tgafile);										// Close The File
        //ShowError(10, 'TGATexture.LoadTGA', False);
        result := false;										// Return False
        exit;
      end;

    i := 0;
    while i < imageSize do
    with textureimage do
    begin
     Tm := ImageData[I+2];
     imageData[i+2] := imageData[i];					// Set The 3rd Byte To The Value In 'temp' (1st Byte Value)
     imageData[i] := Tm;                          // Set The 1st Byte To The Value Of The 3rd Byte

     i := i + bytesPerPixel;
    end;

 end
 else  //COMPRESSED TGA'S
 begin
   PixelCount := textureimage.width * Textureimage.Height;
   CurrentPixel := 0;
   CurrentByte := 0;
   GetMem(ColorBuffer, BytesPerPixel);

   Repeat
      ChunkHeader := 0;
      Ret:=FileName.Read(ChunkHeader, sizeof(gluByte));
      //Ret:=FileRead(tgaFile, ChunkHeader, sizeof(gluByte));
      if Ret = 0 then
      begin
        //ERROR reading Chunk!
        //fileclose(tgafile);										// Close The File
        //ShowError(18, 'TGATexture.LoadTGA', False);
        result := false;										// Return False
        exit;
      end;


      if ChunkHeader < 128 then
      begin
         ChunkHeader := ChunkHeader + 1;
         For Counter := 0 to ChunkHeader-1 do
         begin
            Ret:=Filename.Read(ColorBuffer^, BytesPerPixel);
            //Ret:=fileRead(tgafile, ColorBuffer^, BytesPerPixel);
            if Ret <> BytesPerPixel then
            begin
             //fileclose(tgafile);										// Close The File
             //ShowError(19, 'TGATexture.LoadTGA(ColorBuffer)', False);
             result := false;										// Return False
             exit;
            end;

            Textureimage.imageData[CurrentByte] := (ColorBuffer[2]);
            Textureimage.imageData[CurrentByte+1] := (ColorBuffer[1]);
            Textureimage.imageData[CurrentByte+2] := (ColorBuffer[0]);
            if BytesPerPixel = 4 then
            Textureimage.imageData[CurrentByte+3] := (ColorBuffer[3]);

            CurrentByte := CurrentByte + bytesPerPixel;
            inc(CurrentPixel);
            if CurrentPixel > PixelCount then
            begin
             //fileclose(tgafile);										// Close The File
             //ShowError(18, 'TGATexture.LoadTGA', False);
             result := false;										// Return False
             exit;
            end;
         end;
      end
      else //Chunkheader > 128
      begin
         ChunkHeader := ChunkHeader - 128;
         Ret:=Filename.Read(Colorbuffer^, BytesPerPixel);
         //Ret:=fileRead(tgafile, ColorBuffer^, BytesPerPixel);
         if Ret <> BytesPerPixel then
         begin
          //fileclose(tgafile);										// Close The File
          //ShowError(19, 'TGATexture.LoadTGA(Load RLE)', False);
          result := false;										// Return False
          exit;
         end;
         For Counter := 0 to ChunkHeader do
         begin
            Textureimage.imageData[CurrentByte] := ColorBuffer[2];
            Textureimage.imageData[CurrentByte+1] := ColorBuffer[1];
            Textureimage.imageData[CurrentByte+2] := ColorBuffer[0];
            if BytesPerPixel = 4 then
            Textureimage.imageData[CurrentByte+3] := ColorBuffer[3];

            CurrentByte := CurrentByte + bytesPerPixel;
            inc(CurrentPixel);
         end;

      end;
   Until CurrentPixel >= PixelCount;
 end;
 //fileclose (tgafile);

   if textureimage.bpp = 24 then									// Was The TGA 24 Bits
            gltype := GL_RGB
            else
            gltype := GL_RGBA;										// If So Set T
 											// Close The File
 Texture:=CreateTexture(textureimage.width, textureimage.height, gltype, textureimage.imagedata);

   FreeMem(textureimage.imageData);



end;
{------------------------------------------------------------------}
{  Determines file type and sends to correct function              }
{------------------------------------------------------------------}
function LoadTexture(Filename: String; var Texture : GLuint) : Boolean;
var stream : TFileStream;
begin
  Stream:=TFileStream.Create(Filename,$0000);
  Result:=LoadTextureStream(Stream, ExtractFileExt(filename), Texture);
  Stream.Free;
  //Log.LogStatus('texture id: '+inttostr(Texture),'textures');
end;

{------------------------------------------------------------------}
{  Determines file type and sends to correct function              }
{------------------------------------------------------------------}
function LoadTextureStream(Stream: TStream; Kind: string; var Texture : GLuint) : Boolean;
begin
  Result:=False;
  if Uppercase(Kind) = '.BMP' then
    Result:=LoadBMPTexture(Stream, Texture);
  if Uppercase(Kind) = '.TGA' then
    Result:=LoadTGATexture(Stream, Texture);
  if Uppercase(Kind) = '.FNT' then
    Result:=LoadTGATexture(Stream, Texture);
end;


end.
